#!/usr/bin/Rscript

#work
#source('.../code/gene set/program/construct_programs.R')


#################################################################################
#required libraries
suppressMessages(library(data.table))
library(Matrix)
#######################################################################
# Variables
#######################################################################
root_dir = ".../"

data_dir = paste(root_dir,'working/gene set/',sep='')
####################
#input

wt_path = paste(data_dir,'wt/all_days.txt',sep='')
wt2_path = paste(data_dir,'wt/day_60_v_20.txt',sep='')
ko_wt_path = paste(data_dir,'ko v wt/all_days.txt',sep='')

####################
#output

program_path = paste(data_dir,'program/program.txt',sep='')

#log file
log_path = paste(root_dir,'code/gene set/program/log/construct_programs.log',sep='')

####################
#other

wt_names = c('expressed','day_30_v_20_up','day_60_v_30_up')
wt2_names = c('day_60_v_20_up')
ko_wt_names = c('expressed','day_30_down')
#######################################################################
# functions
#######################################################################
#read gene annotations from MAGMA format file: 2 tab-delimited fields (no field names/header), 1 line per gene-set
#field 1 = annotation name
#field 2 = space (' ') separated list of entrez gene ids
# e.g.
#ann_name1	entrez1 entrez2 entrez3 ...
#ann_name2...
read_MAGMA <- function(next_path) {
	
	tmp = fread(next_path,header = FALSE,col.names = c('ann','entrez'))
	
	tmp_list = lapply(tmp$ann,function(k) unlist(strsplit(tmp[ann == k,entrez],' ',fixed = TRUE)))
	names(tmp_list) = tmp$ann
	
	return(tmp_list)
}
########################################################################################
#extract overlap with downregulated genes & these save gene-sets (& covariates) to file
write_MAGMA <- function(path,data) {
	
	entrez_ids = mapply(function(k) paste(data[[k]],sep='',collapse = ' '),names(data))
	
	magma_table = data.table(ann = names(data), entrez = entrez_ids)
	
	write.table(magma_table,file = path,row.names = FALSE,col.names = FALSE,sep = '\t',quote = FALSE)
	
	cat('Writing gene sets to file',path,'\n\n')
}
##########################################################################
#initialisation
###########################################################################
#log file
sink(file = log_path, append = F)
#########################################################################################################################################
# main - start
#########################################################################################################################################

#read data files & extract gene sets
wt_data = read_MAGMA(wt_path)[wt_names]
wt2_data = read_MAGMA(wt2_path)[wt2_names]
ko_wt_data = read_MAGMA(ko_wt_path)[ko_wt_names]

cat(length(wt_data),'gene-sets read from file:',wt_path,'\n\n')
for (k in names(wt_data)) { cat(length(wt_data[[k]]),'\t',k,'\n')}
cat('\n')

cat(length(wt2_data),'gene-sets read from file:',wt2_path,'\n\n')
for (k in names(wt2_data)) { cat(length(wt2_data[[k]]),'\t',k,'\n')}
cat('\n')

cat(length(ko_wt_data),'gene-sets read from file:',ko_wt_path,'\n\n')
for (k in names(ko_wt_data)) { cat(length(ko_wt_data[[k]]),'\t',k,'\n')}
cat('\n')

##################################
#construct composite groups needed
program_data = list('expressed' = union(ko_wt_data$expressed,wt_data$expressed))

###############################################################
# early increasing: genes up-regulated in wt days 20-30 and 30-60
early_increasing = intersect(wt_data$day_30_v_20_up,wt_data$day_60_v_30_up)

# [-/-] = subset down-regulated in wt v ko at day 30
program_data$early_increasing_ko = intersect(early_increasing,ko_wt_data$day_30_down)

# [+/+] = subset NOT down-regulated in wt v ko at day 30
program_data$early_increasing_wt_only = setdiff(early_increasing,ko_wt_data$day_30_down)

################################################################################
# early stable: genes up-regulated in wt days 20-30 and 20-60, but not 30-60
early_stable = setdiff(intersect(wt_data$day_30_v_20_up,wt2_data$day_60_v_20_up),wt_data$day_60_v_30_up)

# [-/-] = subset down-regulated in wt v ko at day 30
program_data$early_stable_ko = intersect(early_stable,ko_wt_data$day_30_down)

# [+/+] = subset NOT down-regulated in wt v ko at day 30
program_data$early_stable_wt_only = setdiff(early_stable,ko_wt_data$day_30_down)

################################################################################
# early transient: genes up-regulated in wt days 20-30 but not 20-60
early_transient = setdiff(wt_data$day_30_v_20_up,wt2_data$day_60_v_20_up)

# [-/-] = subset down-regulated in wt v ko at day 30
program_data$early_transient_ko = intersect(early_transient,ko_wt_data$day_30_down) 

# [+/+] = subset NOT down-regulated in wt v ko at day 30
program_data$early_transient_wt_only = setdiff(early_transient,ko_wt_data$day_30_down) 

#########################################################
# late: genes up-regulated in wt days 30-60 but not 20-30
program_data$late = setdiff(wt_data$day_60_v_30_up,wt_data$day_30_v_20_up)

# [-/-] = subset down-regulated in wt v ko at day 30
late_ko = intersect(program_data$late,ko_wt_data$day_30_down)

cat(length(late_ko),'out of',length(program_data$late),'[',100*length(late_ko)/length(program_data$late),'%] genes in late set overlap day_30_down_ko\n\n')

####################################################################################

#save gene-sets for analysis
write_MAGMA(program_path,program_data)

cat('Program gene-sets written to file:', program_path,'\n\n')
for (k in names(program_data)) { cat(length(program_data[[k]]),k,'\n')}
cat('\n')
##############################
#stop sending data to log file
sink()